/**
 * A quick-and-dirty circular queue
 *
 * To enable testing, compile with:
 *
 *   cc -DTEST_QUEUE queue.c
 *
 * Use of this code in production is not recommended!
 *
 * This code is hereby granted to the public domain.
 */

#include "queue.h"

/**
 * Initialize a queue
 */
void queue_init(struct queue *q)
{
	q->head = q->tail = 0;
	q->count = 0;
}

/**
 * Add a value to the queue
 *
 * @return -1 if queue is full
 */
int queue_enqueue(struct queue *q, int val)
{
	if (queue_full(q)) return -1;

	q->entry[q->head] = val;
	q->head++;
	q->head %= QUEUE_SIZE;

	q->count++;

	return 0;
}

/**
 * Remove a value from the queue
 *
 * @return -1 if queue is empty
 */
int queue_dequeue(struct queue *q, int *val)
{
	if (queue_empty(q)) return -1;

	*val = q->entry[q->tail];
	q->tail++;
	q->tail %= QUEUE_SIZE;

	q->count--;

	return 0;
}

/**
 * Return true if the queue is full
 */
int queue_full(struct queue *q)
{
	return q->count == QUEUE_SIZE;
}

/**
 * Return true if the queue is empty
 */
int queue_empty(struct queue *q)
{
	return q->count == 0;
}

/**
 * Return the number of items in the queue
 */
int queue_count(struct queue *q)
{
	return q->count;
}

/*-----------------------------------------------------------------------
** test code
*/

#ifdef TEST_QUEUE

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define FUZZ_TEST_ROUNDS (10000000)

void test_queue(int cond, char *str)
{
	if (!cond) {
		fprintf(stderr, "TEST FAIL: %s\n", str); fflush(stderr);
		exit(EXIT_FAILURE);
	}
}

int main(void)
{
	int count;
	int tailval;
	int headval;
	struct queue q1;
	int i;
	int rv;
	int val;
	int enqueue_count, dequeue_count, full_count, empty_count;

	srand(time(NULL));

	queue_init(&q1);

	count = 0;
	tailval = headval = 0;
	enqueue_count = dequeue_count = full_count = empty_count = 0;

	for (i = 0; i < FUZZ_TEST_ROUNDS; i++) {

		// perform a random operation on the queue:
		switch (rand()%2) {
			case 0: // enqueue
				rv = queue_enqueue(&q1, headval);

				// make sure enqueue failed if it was supposed to:
				if (count == QUEUE_SIZE) {
					test_queue(rv == -1, "enqueue 1");
				} else {
					test_queue(rv != -1, "enqueue 2");
				}

				if (rv == 0) {
					enqueue_count++;
					headval++;
					count++;
				}
				break;

			case 1: // dequeue
				rv = queue_dequeue(&q1, &val);

				// make sure dequeue failed if it was supposed to:
				if (count == 0) {
					test_queue(rv == -1, "dequeue 1");
				} else {
					test_queue(rv != -1, "dequeue 2");
				}

				if (rv == 0) {
					dequeue_count++;

					// make sure we dequeued the expected value:
					test_queue(val == tailval, "dequeue val");

					tailval++;
					count--;
				}
				break;
		}

		// make sure queue_full() is reporting properly:
		if (count == QUEUE_SIZE) {
			test_queue(queue_full(&q1), "queue full 1");
			full_count++;
		} else {
			test_queue(!queue_full(&q1), "queue full 2");
		}

		// make sure queue_empty() is reporting properly:
		if (count == 0) {
			test_queue(queue_empty(&q1), "queue empty 1");
			empty_count++;
		} else {
			test_queue(!queue_empty(&q1), "queue empty 2");
		}

		// make sure queue_count() is reporting properly:
		test_queue(queue_count(&q1) == count, "queue count");
	}

	// print summary statistics for fun
	// (and to verify stuff actually happened)
	printf("  Fuzz rounds: %10d\n", FUZZ_TEST_ROUNDS);
	printf("Enqueue count: %10d\n", enqueue_count);
	printf("Dequeue count: %10d\n", dequeue_count);
	printf("   Full count: %10d\n", full_count);
	printf("  Empty count: %10d\n", empty_count);

	return EXIT_SUCCESS;
}

#endif // ifdef TEST_QUEUE

